'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

exports.TokenSymbol = void 0;
(function (TokenSymbol) {
    TokenSymbol[TokenSymbol["DIGIT"] = 1] = "DIGIT";
    TokenSymbol[TokenSymbol["OPEN_PAREN"] = 2] = "OPEN_PAREN";
    TokenSymbol[TokenSymbol["CLOSE_PAREN"] = 3] = "CLOSE_PAREN";
    TokenSymbol[TokenSymbol["PLUS"] = 4] = "PLUS";
    TokenSymbol[TokenSymbol["MINUS"] = 5] = "MINUS";
    TokenSymbol[TokenSymbol["MULTI"] = 6] = "MULTI";
    TokenSymbol[TokenSymbol["DIVIDE"] = 7] = "DIVIDE";
    TokenSymbol[TokenSymbol["END"] = 8] = "END";
    TokenSymbol[TokenSymbol["A"] = -1] = "A";
    TokenSymbol[TokenSymbol["B"] = -2] = "B";
    TokenSymbol[TokenSymbol["C"] = -3] = "C";
    TokenSymbol[TokenSymbol["D"] = -4] = "D";
    TokenSymbol[TokenSymbol["E"] = -5] = "E";
})(exports.TokenSymbol || (exports.TokenSymbol = {}));
const ll1IdxMap = Object.freeze({
    '0': exports.TokenSymbol.DIGIT,
    '1': exports.TokenSymbol.DIGIT,
    '2': exports.TokenSymbol.DIGIT,
    '3': exports.TokenSymbol.DIGIT,
    '4': exports.TokenSymbol.DIGIT,
    '5': exports.TokenSymbol.DIGIT,
    '6': exports.TokenSymbol.DIGIT,
    '7': exports.TokenSymbol.DIGIT,
    '8': exports.TokenSymbol.DIGIT,
    '9': exports.TokenSymbol.DIGIT,
    '(': exports.TokenSymbol.OPEN_PAREN,
    ')': exports.TokenSymbol.CLOSE_PAREN,
    '+': exports.TokenSymbol.PLUS,
    '-': exports.TokenSymbol.MINUS,
    '*': exports.TokenSymbol.MULTI,
    '/': exports.TokenSymbol.DIVIDE,
    $: exports.TokenSymbol.END,
    A: exports.TokenSymbol.A,
    B: exports.TokenSymbol.B,
    C: exports.TokenSymbol.C,
    D: exports.TokenSymbol.D,
    E: exports.TokenSymbol.E,
});
const idx = (c) => ll1IdxMap[c];
const sddTable = [
    'BD',
    '+BD',
    '-BD',
    'CE',
    '0',
    '(A)',
    '+BD',
    '-BD',
    '',
    '*CE',
    '/CE',
    '',
].map(x => x.split('').map(idx));
const MAX_TOKENS = 5;
const MAX_SYMBOLS = 8;
const ll1Table = new Array(MAX_TOKENS + 1);
{
    for (let i = 0; i <= MAX_TOKENS; ++i) {
        ll1Table[i] = new Int8Array(MAX_SYMBOLS + 1).fill(-1);
    }
    ll1Table[1][1] = 0;
    ll1Table[1][2] = 0;
    ll1Table[1][4] = 1;
    ll1Table[1][5] = 2;
    ll1Table[2][1] = 3;
    ll1Table[2][2] = 3;
    ll1Table[3][1] = 4;
    ll1Table[3][2] = 5;
    ll1Table[4][4] = 6;
    ll1Table[4][5] = 7;
    ll1Table[4][3] = 8;
    ll1Table[4][8] = 8;
    ll1Table[5][6] = 9;
    ll1Table[5][7] = 10;
    ll1Table[5][3] = 11;
    ll1Table[5][4] = 11;
    ll1Table[5][5] = 11;
    ll1Table[5][8] = 11;
}

const DOT_CHAR_CODE = 46;
const ZERO_CHAR_CODE = 48;
const NINE_CHAR_CODE = 57;
const integerOperations = {
    ZERO: 0,
    parse: (s, start) => {
        let result = 0;
        let i = start;
        for (; i < s.length; ++i) {
            const c = s.charCodeAt(i);
            if (c < ZERO_CHAR_CODE || c > NINE_CHAR_CODE)
                break;
            result = result * 10 + c - ZERO_CHAR_CODE;
        }
        return [i, result];
    },
    add: (a, b) => a + b,
    subtract: (a, b) => a - b,
    multiply: (a, b) => a * b,
    divide: (a, b) => Math.floor(a / b),
};
const decimalOperations = {
    ZERO: 0,
    parse: (s, start) => {
        let i = start;
        let dotCnt = false;
        for (; i < s.length; ++i) {
            const c = s.charCodeAt(i);
            if (c === DOT_CHAR_CODE) {
                if (dotCnt)
                    return [-1, Number.NaN];
                dotCnt = true;
                continue;
            }
            if (c < ZERO_CHAR_CODE || c > NINE_CHAR_CODE)
                break;
        }
        return [i, Number(s.slice(start, i))];
    },
    add: (a, b) => a + b,
    subtract: (a, b) => a - b,
    multiply: (a, b) => a * b,
    divide: (a, b) => a / b,
};
const bigintOperations = {
    ZERO: 0n,
    parse: (s, start) => {
        let i = start;
        for (; i < s.length; ++i) {
            const c = s.charCodeAt(i);
            if (c < ZERO_CHAR_CODE || c > NINE_CHAR_CODE)
                break;
        }
        return [i, BigInt(s.slice(start, i))];
    },
    add: (a, b) => a + b,
    subtract: (a, b) => a - b,
    multiply: (a, b) => a * b,
    divide: (a, b) => a / b,
};

function createCalculate(operations, resolveExpression) {
    const { ZERO, parse, add, subtract, multiply, divide } = operations;
    return function (rawExpression) {
        const expression = resolveExpression == null ? rawExpression : resolveExpression(rawExpression);
        let cur = 0;
        const result = execute(idx('A'), ZERO, ZERO);
        if (cur === expression.length)
            return result;
        throw new SyntaxError('Not a valid arithmetic expression.');
        function execute(id, syn, inh) {
            if (cur === expression.length) {
                if (id === exports.TokenSymbol.D || id === exports.TokenSymbol.E)
                    return inh;
                throw new SyntaxError('Not a valid arithmetic expression.');
            }
            const id0 = idx(expression[cur]);
            if (id0 === undefined) {
                throw new SyntaxError(`Unrecognized symbol: ${expression[cur]}`);
            }
            if (id === id0) {
                if (id0 === exports.TokenSymbol.DIGIT) {
                    const [nextCur, value] = parse(expression, cur);
                    if (cur >= nextCur) {
                        throw new SyntaxError(`Invalid number`);
                    }
                    cur = nextCur;
                    return value;
                }
                cur += 1;
                return syn;
            }
            if (id > 0) {
                throw new SyntaxError('Not a valid arithmetic expression.');
            }
            const ssdId = ll1Table[-id][id0];
            if (ssdId < 0) {
                throw new SyntaxError('Not a valid arithmetic expression.');
            }
            const tokens = sddTable[ssdId];
            const syn0 = tokens.length > 0 ? execute(tokens[0], ZERO, ZERO) : ZERO;
            switch (ssdId) {
                case 0:
                    return execute(tokens[1], ZERO, syn0);
                case 1: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, val1);
                }
                case 2: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, subtract(ZERO, val1));
                }
                case 3:
                    return execute(tokens[1], ZERO, syn0);
                case 4:
                    return syn0;
                case 5: {
                    const result = execute(tokens[1], ZERO, ZERO);
                    execute(tokens[2], ZERO, ZERO);
                    return result;
                }
                case 6: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, add(inh, val1));
                }
                case 7: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, subtract(inh, val1));
                }
                case 8:
                    return inh;
                case 9: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, multiply(inh, val1));
                }
                case 10: {
                    const val1 = execute(tokens[1], ZERO, ZERO);
                    return execute(tokens[2], ZERO, divide(inh, val1));
                }
                case 11:
                    return inh;
                default:
                    throw new Error('Shit codes!');
            }
        }
    };
}
const calculate = createCalculate(integerOperations, s => s.replace(/[\s]+/g, ''));
const decimalCalculate = createCalculate(decimalOperations, s => s.replace(/[\s]+/g, '').replace(/(^|[^\d.])\./g, '$10.'));
const bigintCalculate = createCalculate(bigintOperations, s => s.replace(/[\s]+/g, ''));

exports.MAX_SYMBOLS = MAX_SYMBOLS;
exports.MAX_TOKENS = MAX_TOKENS;
exports.bigintCalculate = bigintCalculate;
exports.bigintOperations = bigintOperations;
exports.calculate = calculate;
exports.createCalculate = createCalculate;
exports.decimalCalculate = decimalCalculate;
exports.decimalOperations = decimalOperations;
exports["default"] = calculate;
exports.idx = idx;
exports.integerOperations = integerOperations;
exports.ll1IdxMap = ll1IdxMap;
exports.ll1Table = ll1Table;
exports.sddTable = sddTable;
